library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity bitcount is
   port
   (
      din   : in std_logic_vector(255 downto 0);
      sum_1 : out std_logic_vector(8 downto 0)
   );
end;

architecture behave of bitcount is


   function bits2num (inp : std_logic_vector(3 downto 0))
      return integer
   is
      variable res : integer;
   begin
      case inp is
         when "0000" => res := 0;
         when "0001" => res := 1;
         when "0010" => res := 1;
         when "0011" => res := 2;
         when "0100" => res := 1;
         when "0101" => res := 2;
         when "0110" => res := 2;
         when "0111" => res := 3;
         when "1000" => res := 1;
         when "1001" => res := 2;
         when "1010" => res := 2;
         when "1011" => res := 3;
         when "1100" => res := 2;
         when "1101" => res := 3;
         when "1110" => res := 3;
         when "1111" => res := 4;
         when others => res := 0; -- nur fuer Sim
      end case;
      return res;
   end;
   
   type LOOK_UP_TYPE is array (0 to 63) of std_logic_vector(3 downto 0);
   type ADD_STAGE_1_TYPE is array (0 to 31) of std_logic_vector(3 downto 0);
   type ADD_STAGE_2_TYPE is array (0 to 15) of std_logic_vector(4 downto 0);
   type ADD_STAGE_3_TYPE is array (0 to 7)  of std_logic_vector(5 downto 0);
   type ADD_STAGE_4_TYPE is array (0 to 3)  of std_logic_vector(6 downto 0);
   type ADD_STAGE_5_TYPE is array (0 to 1)  of std_logic_vector(7 downto 0);
   
   signal look_up : LOOK_UP_TYPE;
   signal add1 : ADD_STAGE_1_TYPE;
   signal add2 : ADD_STAGE_2_TYPE;
   signal add3 : ADD_STAGE_3_TYPE;
   signal add4 : ADD_STAGE_4_TYPE;
   signal add5 : ADD_STAGE_5_TYPE;

begin
   process(din, look_up, add1, add2, add3, add4, add5)
   begin
      -- LookUp Function
      for i in 0 to 63 loop
         look_up(i) <= conv_std_logic_vector(bits2num( din(4*i+3 downto 4*i) ), 4);
      end loop;
      
      -- 1st Adder Stage
      for k in 0 to ADD_STAGE_1_TYPE'HIGH loop
         add1(k) <= (look_up(2*k) + look_up(2*k+1));
      end loop;
      
      -- 2nd Adder Stage
      for l in 0 to ADD_STAGE_2_TYPE'HIGH loop
         add2(l) <= ('0' & add1(2*l)) + ('0' & add1(2*l+1));
      end loop;
         
      -- 3rd Adder Stage
      for m in 0 to ADD_STAGE_3_TYPE'HIGH loop
         add3(m) <= ('0' & add2(2*m)) + ('0' & add2(2*m+1));
      end loop;
      
      -- 4th Adder Stage
      for n in 0 to ADD_STAGE_4_TYPE'HIGH loop
         add4(n) <= ('0' & add3(2*n)) + ('0' & add3(2*n+1));
      end loop;
      
      -- 5th Adder Stage
      for o in 0 to ADD_STAGE_5_TYPE'HIGH loop
         add5(o) <= ('0' & add4(2*o)) + ('0' & add4(2*o+1));
      end loop;
      
      -- finally
      sum_1 <= (('0' & add5(0)) + ('0' & add5(1)));
      
   end process;
   
         
end;